from collections import OrderedDict
import inspect

try:
    # use Python3 reload
    from imp import reload

except:
    # we are on Python2
    pass

class Registry(OrderedDict):

    def register_decorator_factory(self, **kwargs):
        """ 
            Return an actual decorator for registering objects into registry
        """ 
        name = kwargs.get('name')
        def decorator(decorated):
            self.register_func(obj=decorated, name=name)
            return decorated
        return decorator

    def register(self, obj=None, name=None, **kwargs):
        """
            Use this method as a decorator on class/funciton you want to register:

            @registry.register(name="test")
            class Test:
                pass

            :param:obj: An object to register in the registry
            :param:name: The name of the object to register. If none, the obj class name will be used


        """
        if obj is None:
            return self.register_decorator_factory(obj=obj, name=name, **kwargs)
        else:
            self.register_func(obj=obj, name=name, **kwargs)
            return obj

    def get_object_name(self, obj):
        """
            Return a name from an element (object, class, function...)
        """
        if callable(obj):
            return obj.__name__

        elif inspect.isclass(obj):
            return obj.__class__.__name__

        else:
            raise ValueError("Cannot deduce name from given object ({0}). Please user registry.register() with a 'name' argument.".format(obj))

    def validate(self, obj):
        """
            Called before registering a new value into the registry
            Override this method if you want to restrict what type of data cna be registered
        """
        return True

    def register_func(self, obj, name=None, **kwargs):
        """
            Register an object, class, function... into the registry
        """
        if self.validate(obj):
            if name is None:
                name = self.get_object_name(obj)
            self[name] = obj
        else:
            raise ValueError("{0} (type: {0.__class__}) is not a valid value for {1} registry".format(obj, self.__class__))

    def autodiscover(self, apps, force_reload=True):
        """
            Iterate throught every installed apps, trying to import `look_into` package
            :param apps: an iterable of string, refering to python modules the registry will try to import via autodiscover
        """
        for app in apps:
            try:
                package = '{0}.{1}'.format(app, self.look_into) # try to import self.package inside current app
                #print(package)
                module = __import__(package)
                if force_reload:
                    reload(module)
            except ImportError:
                # Module does not exist 
                pass


class MetaRegistry(Registry):
    """
        Keep a reference to all registries
    """
    look_into = "registries"

    def autodiscover(self, apps, cascade=True, **kwargs):
        """
            :param cascade: If true, will trigger autodiscover on discovered registries
        """
        super(MetaRegistry, self).autodiscover(apps, **kwargs)
        if cascade:
            self.autodiscover_registries(apps)

    def autodiscover_registries(self, apps):
        for key, registry in self.items():
            registry.autodiscover(apps)
            
meta_registry = MetaRegistry()


import os, sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from persisting_theory import Registry

class CallbacksRegistry(Registry):
    """
        Allow your apps to register callbacks
    """
    # the package where the registry will try to find callbacks in each app
    look_into = "callbacks_registry"

callbacks_registry = CallbacksRegistry()

APPS = (
    'app1',
    'app2',
)
# Trigger autodiscovering process
callbacks_registry.autodiscover(APPS)


import os, sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
import unittest 
import test_registries
import registries

registries.meta_registry.look_into = "test_registries"

TEST_APPS = (
    "app1",
    "app2",
)

class RegistryTest(unittest.TestCase):

    def test_can_infer_name_from_class_function_and_instance(self):
        registry = registries.Registry()

        def something():
            pass

        class MyClass:
            pass

        self.assertEqual(registry.get_object_name(something), "something")
        self.assertEqual(registry.get_object_name(MyClass), "MyClass")

        with self.assertRaises(ValueError):
            self.assertEqual(registry.get_object_name(MyClass()), "MyClass")

    def test_can_register_data_to_registry(self):

        data = "something"
        registry = registries.Registry()
        registry.register(data, name="key")

        self.assertEqual(len(registry), 1)
        self.assertEqual(registry.get("key"), data)

    def test_can_restric_registered_data(self):

        class RestrictedRegistry(registries.Registry):
            def validate(self, obj):
                """Only accept integer values"""
                return isinstance(obj, int)

        registry = RestrictedRegistry()

        registry.register(12, name="twelve")
        with self.assertRaises(ValueError):
            registry.register("not an int", name="not an int")


    def test_can_register_class_and_function_via_decorator(self):
        registry = registries.Registry()

        @registry.register
        class ToRegister:
            pass

        self.assertEqual(registry.get('ToRegister'), ToRegister)

        @registry.register
        def something():
            pass

        self.assertEqual(registry.get('something'), something)
            
    def test_can_register_via_decorator_using_custom_name(self):
        registry = registries.Registry()

        @registry.register(name="custom_name")
        def something():
            pass

        self.assertEqual(registry.get('custom_name'), something)

    def test_registry_can_autodiscover(self):

        registry = test_registries.awesome_people
        registry.autodiscover(apps=TEST_APPS)

        self.assertEqual(len(registry), 2)
        self.assertNotEqual(registry.get('AlainDamasio', None), None)
        self.assertNotEqual(registry.get('FrederikPeeters', None), None)

        registry.clear()

    def test_meta_registry_can_autodiscovering_registries_and_trigger_their_autodiscover_method(self):

        registry = registries.meta_registry
        registry.autodiscover(apps=TEST_APPS)

        self.assertEqual(len(registry), 2)
        self.assertEqual(registry.get('awesome_people'), test_registries.awesome_people)
        self.assertEqual(registry.get('vegetable_registry'), test_registries.vegetable_registry)

        registry = test_registries.awesome_people
        self.assertEqual(len(registry), 2)
        self.assertNotEqual(registry.get('AlainDamasio', None), None)
        self.assertNotEqual(registry.get('FrederikPeeters', None), None)

        registry = test_registries.vegetable_registry
        self.assertEqual(len(registry), 2)
        self.assertNotEqual(registry.get('Potato', None), None)
        self.assertNotEqual(registry.get('Ketchup', None), None)


if __name__ == '__main__':
    unittest.main()


import os
from setuptools import setup, find_packages
import persisting_theory

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='persisting-theory',
    version=persisting_theory.__version__,
    packages=find_packages(),
    include_package_data=True,
    license='BSD',  # example license
    description='A python package to build registries that can autodiscover values accross your project components',
    long_description=README,
    url='http://code.eliotberriot.com/eliotberriot/persisting-theory',
    author='Eliot Berriot',
    author_email='contact@eliotberriot.com',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
    ],
)


from registries import Registry, meta_registry

class AwesomePeopleRegistry(Registry):
    look_into = "awesome_people"

awesome_people = AwesomePeopleRegistry()
meta_registry.register(awesome_people, name="awesome_people")

class VegetableRegistry(Registry):
    look_into = "vegetables"

vegetable_registry = VegetableRegistry()
meta_registry.register(vegetable_registry, name="vegetable_registry")


from registries import callbacks_registry
    
APPS = (
    'app1',
    'app2',
)

# Trigger autodiscovering process
callbacks_registry.autodiscover(APPS)

for callback in callbacks_registry.values():
    callback()

    # Wouf
    # Meow